{$X+}

(***************************************)
(* Copyright (C) 2007, SHIRAISHI Kazuo *)
(***************************************)

unit compiler;


interface
uses   mdiedit, struct;


procedure RunNormal(EditForm:TEditForm);
procedure RunStep(EditForm:TEditForm);

var
   changedir:procedure;
   setIndentOption:procedure;
   setOperation:procedure;
   setRunOption:procedure;
   setRunOption2:procedure;
   setDebug:procedure;
   setLineEndMarker:procedure;
   setRefferingPath:procedure;
var
   ProgramOnRunning:string='';

implementation

uses  QForms, Classes, QDialogs, SysUtils, QComCtrls, QControls, QGraphics, QStdCtrls,
     mdiframe,paintfrm,textfrm,tracefrm,  listcoll, base,arithmet,
     myutils,rational,texthand,textfile,express,statemen,graphic,graphsys,
     optiondg,debugdg,helpctex,math2sub,charinp,sconsts,vstack,memman,merge;

type
  TBackUp=class(TThread)
       text1:ansistring;
       fName:ansistring;
    constructor create(const s,f:AnsiString);
    procedure execute;override;
  end;

constructor TBackUp.create(const s,f:AnsiString);
begin
   inherited create(false);
   text1:=s;
   FName:=f;
end;

procedure TBackUp.execute;
var
  t:Text;
begin
  assignFile(t,fname);
  try
    rewrite(t);
    write(t,text1);
    close(t)
  except
  end;
end;


var
    ReCompile:boolean=false;

function compile:boolean;
begin
 Screen.cursor:=crHourGlass;
 compile:=true;
 recompile:=false;
 try
    struct.compile;
 except
   on E:Exception do
      begin
          currentprogram.freeall;
          compile:=false;
          if E is EReCompile then
             recompile:=true
          else if not (E is SyntaxError) then
             ShowMessage(s_internalErrorCompiling +EOL+ E.message  +EOL+ Contact);
      end;
 end;
 Screen.cursor:=crDefault;
end;

function CompilePrg(EditForm:TEditForm):boolean;
var
   Cpos:TCaretPos;
   svUseTranscendentalFunction:boolean;
begin
   svUseTranscendentalFunction := UseTranscendentalFunction;

   EditForm.Memo1.lines.BeginUpdate;

   extype:=0;
   InitSeed;

   GraphMode:=false;
   textMode:=false;
   UseCharInput:=false;

   ReCompile:=false;
   repeat

       SetPrecisionMode(InitialPrecisionMode,true);

       indent:=-1;
       USEnest:=0;

       DoStack:=TList.Create;
       ForStack:=TList.create;
       WhenStack:=TList.create;
       WhenStack.add(nil);
       WhenUseStack:=TList.create;
       WhenUseStack.add(nil);

       result:=false;
       pass:=1;

       try
         if compile then
            begin
               pass:=2;
               currentprogram.deletestatements;
               currentprogram.VarTablesRebuild;
               currentprogram.ShareVarTableGetVar;
               result:=compile;
            end;
       finally
         DoStack.Free;
         ForStack.free;
         WhenStack.free;
         WhenUseStack.free;

         KeyWordTablesFreeAll;
         Pass:=0;
       end;
   until recompile=false;

   EditForm.Memo1.lines.EndUpdate;
   //  顼ɽ
   if result=false then
         begin
           SelectLine(EditForm.memo1,exline);
           MessageDlg(s_SyntaxError+#10+statusmes.murgeWithOR, mtWarning, [mbOK]
                                             ,HelpContext);
           CPos.line:=exline;
           CPos.col:=expos+exinsertcount-1;
           with EditForm.memo1 do
              CaretPos:=cpos;
         end;

   UseTranscendentalFunction := svUseTranscendentalFunction

end;

procedure RunPrg(EditForm:TEditForm);
var
   mes:string;
   CurDir:String;
   hc:integer;
   svSelStart:integer;
   svCapture:TControl;
begin
   extype:=0;
   CurrentOperation:=nil;
   statusmes.clear;
   DebugDlg.init;

   svCapture:=Mouse.Capture;
   Mouse.Capture:=FrameForm;

  if not GraphMode  then
          PaintForm.WindowState:=wsMinimized;

  CurDir:=ExtractFilePath(EditForm.Caption);
  if (CurDir<>'') and (Curdir<>GetCurrentDir) then
     SetCurrentDir(CurDir);

  console:=TConsole.create;
  PConsole.ttext:=console;           //2008.11.3
  LocalPrinter:=TLocalPrinter.create;
  TextForm.Caption:=ChangeFileExt(EditForm.Caption,'.txt');

  TraceForm.Memo1.lines.text:='';
  TraceForm.Caption:=ChangeFileExt(EditForm.Caption,'.log');
  TraceForm.setReadOnly(true);

  if BreakFlags.TraceMode then
     begin
         TraceForm.Visible:=true;
         TraceForm.WindowState:=wsNormal;
         TraceForm.BringToFront;
     end
  else
     TraceForm.Visible:=false;  //WindowState:=wsMinimized;


  PaintForm.initial;
  InitGraphics;
  if graphmode and  (nextGraphmode=ScreenBitmapMode) then
       begin
          PaintForm.Caption:=ChangeFileExt(EditForm.Caption,'.bmp');
          PaintForm.WindowState:=wsNormal;
          PaintForm.BringToFront;
       end ;

  if UseCharInput then
    begin
       charinput.init;
       CharInput.Show
    end;

   Set8087CW(controlword);
   asm
      mov initialESP,esp
   end;

   hc:=0;

  try
     try
        CurrentProgram.RunModules ;
     finally
        currentprogram.ShareVarTableFreeVar;
        currentprogram.deletestatements;
        currentprogram.freeall;
        console.free;
        PConsole.ttext:=nil;        //2008.11.3
        LocalPrinter.free;
        EditForm.memo1.SelLength:=0;
        PaintForm.Repaint;
        charinput.hide;
     end;
  except
  on EExtype do
    begin
      hc:=0;
      case extype mod 100000 of
           0      : mes:='' ;
        1001      : mes:=s_Extype1001;
        1002      : mes:=s_Extype1002;
        1003      : mes:=s_Extype1003;
        1006      : mes:=s_Extype1006;
        1007      : mes:=s_Extype1007;
        1008      : mes:=s_Extype1008;
        1050..1106: mes:=s_Extype1050;
        1004..1005,
        1009..1049,
        1107..1999: mes:=s_Extype1000;

        2001      : begin mes:=s_Extype2001; hc:=IDH_ARRAY end;
        3000      : mes:=s_Extype3000;
        3001      : mes:=s_Extype3001;
        3002      : mes:=s_Extype3002;
        3003      : mes:=s_Extype3003;
        3004..3008: mes:=s_Extype3004;
        3009      : MES:=s_Extype3009;
        4000..4299:
               begin
                    mes:=s_Extype3004;
                    case extype mod 100000 of
                         4004: mes:=mes + '(SIZE)';
                         4005: mes:=mes + '(TAB)';
                         4008: mes:=mes + '(LBOUND)';
                         4009: mes:=mes + '(UBOUND)';
                         4010: mes:=mes + '(REPEAT$)';
                        else
                    end;
               end;
        5001,5002 : begin mes:=s_Extype5001; hc:=IDH_MAT  end;
        6001..6402: mes:=s_Extype6001;
        7001      : mes:=s_Extype7001;
        7003      : mes:=s_Extype7003;
        7004      : mes:=s_Extype7004;
        7101      : mes:=s_Extype7101;
        7102      : mes:=s_Extype7102;
        7103      : mes:=s_Extype7103;
        7301      : begin mes:=s_EXtype7301; hc:=IDH_ERASE end;
        7302      : mes:=s_EXtype7302;
        7303      : mes:=s_EXtype7303;
        7305      : mes:=s_Extype7305;
        7308      : mes:=s_Extype7308;
        7317      : mes:=s_Extype7317;
        7318      : mes:=s_Extype7318;

        7005..7100,7104..7300,7311..7316,7320..7402
                  : mes:=s_Extype7000;
          8001      : mes:=s_Extype8001;
          8011      : mes:=s_Extype8011;
          8012      : mes:=s_Extype8012;
          8013      : mes:=s_Extype8013;
          8101      : mes:=s_Extype8101;
          8002,8003,8102,8103: mes:=s_Extype8002;
          8105      : mes:=s_Extype8105;
          8120      : mes:=s_Extype8120;
          8201      : begin mes:=s_Extype8201; hc:=IDH_PRINT_USING end;
          8202      : begin mes:=s_Extype8202; hc:=IDH_PRINT_USING end;
          8401      : mes:=s_Extype8401;
          8402      : mes:=s_Extype8402;
          9000      : mes:=s_Extype9000;
          9002      : mes:=s_Extype9002;
          9003      : mes:=s_Extype9003;
          9004      : mes:=s_Extype9004;
          9102      : mes:=s_Extype9102;
          10002     : mes:=s_Extype10002;
          10004     : mes:=s_Extype10004;
          11004     : begin mes:=s_Extype11004; hc:=IDH_SELECT end;
          11051     : begin mes:=s_Extype11051; hc:=IDH_WINDOW end;
          12004     : mes:=s_Extype12004;
        outofmemory : mes:=s_OutoOfMemory;
        virtualStackOverflow:begin mes:=s_VStackOverflow; hc:=IDH_STACK_LIMIT end;
        stackoverflow:begin mes:=s_StackOverflow; hc:=IDH_STACK_LIMIT; end;
        ArraySizeOverflow: begin mes:=s_ArraySizeOverflow; hc:=IDH_LIMIT end;
        TextOverFlow: mes:=s_OutputOverflow;
        RToNOverflow:begin mes:=S_RToNOverflow; hc:=IDH_RATIONAL end;
        systemErr   : mes:='system error';
        else          mes:=''   ;
      end;
    end;
    on E:EStackOverflow do ShowMessage(E.Message);
    on E:EOutOfMemory do ShowMessage(E.Message);
    on E:Exception do
           ShowMessage(s_InternalErrorRunning +EOL+ E.message + EOL+ Contact);

  end;

  statusmes.add(mes);
  With EditForm.memo1 do
     begin
        //LockWindowUpdate(Handle);
        svSelStart:=selStart;
        SelectAll;
        //SelAttributes:=DefAttributes;
        SelStart:=svSelstart;
        SelLength:=0;
        //LockWindowUpdate(0);
     end;

  //EditForm.memo1.HideSelection:=false;
  if extype<>0 then
       begin
          SelectLine(EditForm.memo1,exline);
          EditForm.Show;
          FrameForm.BringToFront;
          mes:='EXTYPE '+strint(extype mod 100000) +EOL+ statusmes.murge ;
          if hc=0 then
             MessageDlg(mes, mtError, [mbOK], 0)
          else
             MessageDlg(mes, mtError, [mbOK], hc) ;

          if extype>0 then
          with DebugDlg do
             begin
               RadioGroup1.visible:=false;
               Execute;
             end;

          statusmes.clear ;
       end;

    MyGraphSys.Finish;

    Mouse.Capture:=svCapture;
end;



function CompileAndRun(EditForm:TEditForm):boolean; //compileTrue;
begin
   Default8087CW:=controlword;
   Set8087CW(controlword);
   InitMemory;
   MemoryManInit;
   InitRational;

   HelpContext:=0;
   StatusMes.clear;
   FrameForm.StatusBar1.Panels[1].text:=s_OnCompiling;
   //FrameForm.StatusBar1.Panels[1].Bevel:=pbLowered;
   FrameForm.StatusBar1.update;

  result:=false;
  if CompilePrg(EditForm) then
      begin
           result:=true;
            if MixedArithmetic then
                FrameForm.StatusBar1.Panels[1].text:=''
           else
                FrameForm.StatusBar1.Panels[1].text:=
                                       precisionText[MainProgram.arithmetic];
           ProgramOnRunning:= ChangeFileExt(EditForm.Caption,'')+s_OnRuunnig;
           FrameForm.StatusBar1.Panels[3].text:=ProgramOnRunning;
           FrameForm.Caption:=AppTitle + s_OnRuunnig;
           FrameForm.StatusBar1.update;
           EditForm.Memo1.Modified:=false;
           try
              RunPrg(EditForm);
           except
            on E:Exception do
               ShowMessage(E.message );
           end;
           FrameForm.Caption:=AppTitle;
     end;
   //FrameForm.StatusBar1.Panels[1].Bevel:=pbNone;
   FrameForm.StatusBar1.Panels[1].text:='';
   FrameForm.StatusBar1.Panels[3].text:=statusBarMems3;
end;

procedure SetExecutingNow(s:boolean; EditForm:TEditForm);
var
   i:integer;
begin
   ExecutingNow:=s;
   with FrameForm do
      begin
        TBRun.enabled:=not s;
        TBStep.enabled:=not s;
        TBBreak.enabled:=s;
        TBCut.enabled:=not s;
        TBPaste.enabled:=not s;
        TBUndo.enabled:=not s;
        if not permitMicrosoft then
           begin
              TBDecimal.enabled:=not s;
              TBHighPrecision.enabled:=not s;
              TBBinary.enabled:=not s;
              TBComplex.enabled:=not s;
              TBRational.enabled:=not s;
              TBDeg.enabled:=not s;
           end;
        Option1.enabled:=not s;
        for i:=0 to  MdiChildCount-1 do
           begin
               if (MdiChildren[i] is TEditForm) then
                  with MdiChildren[i] as TEditForm do
                    begin
                        Run2.enabled:=not s;
                        Break1.enabled:=s;
                        Step1.enabled:=not s;
                        Exit1.enabled:=not s;
                        PopUpRun1.enabled:=not s;
                        PopUpBreak1.enabled:=s;
                        PopUpStep1.enabled:=not s;
                    end
           end;
      end;
   with EditForm do begin
       Close1.enabled:=not s;
       merge1.enabled:=not s;
       Cut1.enabled:=not s;
       Paste1.enabled:=not s;
       Delete1.enabled:=not s;
       Undo1.enabled:=not s;
       Repalce1.enabled:=not s;
       deleteLabelNumber1.enabled:=not s;
       addLabelNumber1.enabled:=not s;
       CaseChange1.enabled:=not s;
       ToolBox1.enabled:=not s;
       Commentize1.Enabled:=not s;
       Uncommentize1.Enabled:=not s;
       Memo1.ReadOnly:=s;
   end;
   with TextForm  do begin
       Break1.enabled:=s;
       File1.enabled:=not s;
       Edit1.enabled:=not s;
       option1.enabled:=not s;
   end;
   with TraceForm  do begin
       Break1.enabled:=s;
       File1.enabled:=not s;
       Edit1.enabled:=not s;
       option1.enabled:=not s;
   end;
   with PaintForm do begin
       break1.enabled:=s;
       File1.enabled:=not s;
       Edit1.enabled:=not s;
       option1.enabled:=not s;
   end;
end;

function appRun(EditForm:TEditForm):boolean;
var
   svModified:boolean;
   BackUp1:TBackUp;
   BkFile:ansistring;
   svKeepText,svKeepGraphic:boolean;
begin
   if TextHand.memo<>nil then exit;   {ɻ}
   TextHand.memo:=EditForm.memo1;
   ChainFile:='';


   BkFile:=changefileext(IniFileName,'.bak');
   if NoBackUp then
      BackUp1:=nil
   else
      BackUp1:=TBackUp.create(EditForm.memo1.lines.text,bkFile);

   MergedLineNumber:=-1;
      
   SetExecutingNow(True,EditForm);
   svModified:=EditForm.Memo1.Modified;

   if permitMicrosoft then
      InitialPrecisionMode:=PrecisionNative
   else
      InitialPrecisionMode:=InitialPrecisionMode0;
   initialCharacterByte:=initialCharacterByte0;

   nextGraphMode:=ScreenBitMapMode;   // LinuxǤΤ

   IdleImmediately;
   try
       result:=CompileAndRun(EditForm);
   except
       on E:Exception do
          ShowMessage(s_InternalErrorCompiling +EOL+ E.message +EOL+ Contact);
   end;

   BackUp1.free;
   SetExecutingNow(False,EditForm);

   if MergedLineNumber>0 then RemoveMergedText;

   EditForm.memo1.Modified:=svModified;
   TextHand.memo:=nil;
   if not NoBackUp then
      SysUtils.DeleteFile(BkFile);

   if ChainFile<>'' then
      begin
        svKeepText:=KeepText;
        svkeepGraphic:=KeepGraphic;
        KeepText:=true;
        KeepGraphic:=true;
        If (FrameForm.MDIChildCount>=10) and not EditForm.Memo1.Modified then
          begin
            EditForm.Close;
            EditForm:=nil;
          end;
        if AppRun(FrameForm.OpenTextFile(ChainFile)) then
           if EditForm<>nil then
              EditForm.BringToFront;

        KeepText:=svKeepText;
        KeepGraphic:=svKeepGraphic;
      end;


end;

procedure RunNormal(EditForm:TEditForm);
begin
    BreakFlags.LongFlag:=false;
    bkDirective:=bkStep;
    if appRun(EditForm) and (extype=0) then
      begin
        if textmode then
          with TextForm do
             begin
               WindowState:=wsNormal;
               BringToFront;
             end;
        if graphMode and  (nextGraphmode=ScreenBitmapMode) then
          with PaintForm do
             begin
               WindowState:=wsNormal;
               BringToFront;
             end;
      end;
end;

procedure RunStep(EditForm:TEditForm);
begin
    BreakFlags.TraceChannelPlus1:=0;
    BreakFlags.TraceMode:=true;
    CtrlBreakHit:=true;
    bkDirective:=bkstep;
    appRun(EditForm);
end;



procedure noefect;far;  
begin
end;

function dummy:TStatement;far;
begin
   dummy:=nil
end;


begin
  setIndentOption:=noefect;
   setoperation:=noefect;
   setRunOption:=noefect;
   setRunOption2:=noefect;
   setDebug:=noefect;
   changedir:=noefect;
   setLineEndMarker:=noefect;
   setRefferingPath:=noefect;

end.
